from typing import Iterable, List, Type, Optional
import numpy as np
from scipy.optimize import fmin_l_bfgs_b
import logging

from .base_classes import SurrogateModel, AcquisitionFunction, \
    ScoringFunction, LocalOptimizer, OutputSurrogateModel
from ..datatypes.common import Candidate
from ..datatypes.tuning_job_state import TuningJobState

logger = logging.getLogger(__name__)


class IndependentThompsonSampling(ScoringFunction):
    """
    Note: This is not Thompson sampling, but rather a variant called
    "independent Thompson sampling", where means and variances are drawn
    from the marginal rather than the joint distribution. This is cheap,
    but incorrect.

    """
    def __init__(
            self, model: SurrogateModel,
            random_state: Optional[np.random.RandomState] = None):
        self.model = model
        if random_state is None:
            random_state = np.random.RandomState(31415629)
        self.random_state = random_state

    def score(self, candidates: Iterable[Candidate],
              model: Optional[SurrogateModel] = None) -> List[float]:
        if model is None:
            model = self.model
        predictions_list = model.predict_candidates(candidates)
        scores = []
        # If the model supports fantasizing, posterior_means is a matrix. In
        # that case, samples are drawn for every column, then averaged (why
        # we need np.mean)
        for predictions in predictions_list:
            posterior_means = predictions['mean']
            posterior_stds = predictions['std']
            new_score = [
                np.mean(self.random_state.normal(m, s))
                for m, s in zip(posterior_means, posterior_stds)]
            scores.append(new_score)
        return list(np.mean(np.array(scores), axis=0))


class LBFGSOptimizeAcquisition(LocalOptimizer):
    def __init__(self, state: TuningJobState, model: OutputSurrogateModel,
                 acquisition_function_class: Type[AcquisitionFunction], active_metric: str = None):
        super().__init__(state, model, acquisition_function_class, active_metric)
        # Number criterion evaluations in last recent optimize call
        self.num_evaluations = None

    def optimize(self, candidate: Candidate,
                 model: Optional[SurrogateModel] = None) -> Candidate:
        # Before local minimization, the model for this state_id should have been fitted.
        if model is None:
            model = self.model
        state = self.state
        acquisition_function = self.acquisition_function_class(model, self.active_metric)

        x0 = state.hp_ranges.to_ndarray(candidate)
        bounds = state.hp_ranges.get_ndarray_bounds()
        n_evaluations = [0]  # wrapped in list to allow access from function

        # unwrap 2d arrays
        def f_df(x):
            n_evaluations[0] += 1
            return acquisition_function.compute_acq_with_gradient(x)

        res = fmin_l_bfgs_b(f_df, x0=x0, bounds=bounds, maxiter=1000)
        self.num_evaluations = n_evaluations[0]
        if res[2]['task'] == b'ABNORMAL_TERMINATION_IN_LNSRCH':
            # this condition was copied from the old GPyOpt code
            logger.warning(
                f"ABNORMAL_TERMINATION_IN_LNSRCH in lbfgs after {n_evaluations[0]} evaluations, "
                "returning original candidate"
            )
            return candidate  # returning original candidate
        else:
            # Clip to avoid situation where result is small epsilon out of bounds
            a_min, a_max = zip(*bounds)
            optimized_x = np.clip(res[0], a_min, a_max)
            # Make sure the above clipping does really just fix numerical rounding issues in LBFGS
            # if any bigger change was made there is a bug and we want to throw an exception
            assert np.linalg.norm(res[0] - optimized_x) < 1e-6, (res[0], optimized_x, bounds)
            result = state.hp_ranges.from_ndarray(optimized_x.flatten())
            return result


class NoOptimization(LocalOptimizer):
    def optimize(self, candidate: Candidate,
                 model: Optional[SurrogateModel]=None) -> Candidate:
        return candidate



import numpy as np


def sanitize_array(array):
    """
    Replace NaN and Inf (there should not be any!)
    :param array:
    :return:
    """
    a = np.ravel(array)
    maxi = np.nanmax(a[np.isfinite(a)])
    mini = np.nanmin(a[np.isfinite(a)])
    array[array == float('inf')] = maxi
    array[array == float('-inf')] = mini
    mid = (maxi + mini) / 2
    array[np.isnan(array)] = mid
    return array



# TODO: Standardize / unify this code with ag.save()
import os, pickle, tempfile, logging, boto3

from ...utils import s3_utils
from ...utils import compression_utils

logger = logging.getLogger(__name__)

compression_fn_map = compression_utils.get_compression_map()

# TODO: object -> obj?
def save(path, object, format=None, verbose=True, **kwargs):
    compression_fn = kwargs.get('compression_fn', None)
    compression_fn_kwargs = kwargs.get('compression_fn_kwargs', None)

    if compression_fn in compression_fn_map:
        validated_path = compression_utils.get_validated_path(path, compression_fn)
    else:
        raise ValueError(f'compression_fn={compression_fn} is not a valid compression_fn. Valid values: {compression_fn_map.keys()}')

    pickle_fn = lambda o, buffer: pickle.dump(o, buffer, protocol=4)
    save_with_fn(validated_path, object, pickle_fn, format=format, verbose=verbose, compression_fn=compression_fn,
                 compression_fn_kwargs=compression_fn_kwargs)

def save_with_fn(path, object, pickle_fn, format=None, verbose=True, compression_fn=None, compression_fn_kwargs=None):
    if verbose:
        logger.log(15, 'Saving '+str(path))
    if s3_utils.is_s3_url(path):
        format = 's3'
    if format == 's3':
        save_s3(path, object, pickle_fn, verbose=verbose)
    else:
        os.makedirs(os.path.dirname(path), exist_ok=True)

        if compression_fn_kwargs is None:
            compression_fn_kwargs = {}

        with compression_fn_map[compression_fn]['open'](path, 'wb', **compression_fn_kwargs) as fout:
            pickle_fn(object, fout)

def save_s3(path: str, obj, pickle_fn, verbose=True):
    if verbose:
        logger.info(f'save object to {path}')
    with tempfile.TemporaryFile() as f:
        pickle_fn(obj, f)
        f.flush()
        f.seek(0)

        bucket, key = s3_utils.s3_path_to_bucket_prefix(path)
        s3_client = boto3.client('s3')
        try:
            config = boto3.s3.transfer.TransferConfig()   # enable multipart uploading for files larger than 8MB
            response = s3_client.upload_fileobj(f, bucket, key, Config=config)
        except:
            logger.error('Failed to save object to s3')
            raise



from typing import Set
import numpy as np

from .gp_fifo_searcher import GPFIFOSearcher, map_reward, MapReward, \
    DEFAULT_INITIAL_SCORING, SUPPORTED_INITIAL_SCORING
from .constrained_gp_fifo_searcher import ConstrainedGPFIFOSearcher
from .gp_multifidelity_searcher import GPMultiFidelitySearcher, \
    resource_for_acquisition_bohb, resource_for_acquisition_first_milestone
from .model_factories import resource_kernel_factory
from ..datatypes.hp_ranges_cs import HyperparameterRanges_CS
from ..datatypes.hp_ranges import HyperparameterRanges
from ..gpautograd.constants import OptimizationConfig, DEFAULT_OPTIMIZATION_CONFIG
from ..gpautograd.gp_regression import GaussianProcessRegression
from ..gpautograd.kernel import Matern52
from ..gpautograd.mean import ScalarMeanFunction
from ..models.gpmodel_skipopt import SkipNoMaxResourcePredicate, SkipPeriodicallyPredicate
from ..models.gp_model import GPModelArgs
from ..models.meanstd_acqfunc_impl import EIAcquisitionFunction, CEIAcquisitionFunction
from ..tuning_algorithms.base_classes import DEFAULT_METRIC, DEFAULT_CONSTRAINT_METRIC
from ..tuning_algorithms.defaults import DEFAULT_NUM_INITIAL_CANDIDATES, DEFAULT_NUM_INITIAL_RANDOM_EVALUATIONS
from ..utils.debug_log import DebugLogPrinter
from ..utils.simple_profiler import SimpleProfiler
from ....utils.default_arguments import Integer, Categorical, Boolean, Float

__all__ = ['gp_fifo_searcher_factory',
           'gp_multifidelity_searcher_factory',
           'gp_fifo_searcher_defaults',
           'gp_multifidelity_searcher_defaults']


def _create_gp_and_model_args(
        hp_ranges_cs, active_metric, max_metric_value=None, is_hyperband=False, **kwargs):
    opt_warmstart = kwargs.get('opt_warmstart', False)
    random_seed = kwargs.get('random_seed', 31415927)
    kernel = Matern52(dimension=hp_ranges_cs.ndarray_size(), ARD=True)
    mean = ScalarMeanFunction()
    if is_hyperband:
        kernel, mean = resource_kernel_factory(
            kwargs['gp_resource_kernel'],
            kernel_x=kernel, mean_x=mean,
            max_metric_value=max_metric_value)
    optimization_config = OptimizationConfig(
        lbfgs_tol=DEFAULT_OPTIMIZATION_CONFIG.lbfgs_tol,
        lbfgs_maxiter=kwargs['opt_maxiter'],
        verbose=kwargs['opt_verbose'],
        n_starts=kwargs['opt_nstarts'])
    gpmodel = GaussianProcessRegression(
        kernel=kernel, mean=mean,
        optimization_config=optimization_config,
        fit_reset_params=not opt_warmstart)
    model_args = GPModelArgs(
        num_fantasy_samples=kwargs['num_fantasy_samples'],
        random_seed=random_seed,
        active_metric=active_metric,
        normalize_targets=True)
    return gpmodel, model_args


def _create_common_objects(**kwargs):
    scheduler = kwargs['scheduler']
    config_space = kwargs['configspace']
    is_hyperband = scheduler.startswith('hyperband')
    if kwargs.get('debug_use_hyperparameter_ranges', False):
        assert isinstance(config_space, HyperparameterRanges)
        assert not is_hyperband, \
            "Cannot use debug_use_hyperparameter_ranges with Hyperband scheduling"
        hp_ranges_cs = config_space
    else:
        import ConfigSpace as CS
        assert isinstance(config_space, CS.ConfigurationSpace)
        hp_ranges_cs = HyperparameterRanges_CS(config_space)
    # Note: This base random seed is used to create different random seeds for
    # each BO get_config call internally
    random_seed = kwargs.get('random_seed', 31415927)
    # Skip optimization predicate for GP surrogate model
    if kwargs.get('opt_skip_num_max_resource', False) and is_hyperband:
        skip_optimization = SkipNoMaxResourcePredicate(
            init_length=kwargs['opt_skip_init_length'],
            resource_attr_name=kwargs['resource_attribute'],
            max_resource=kwargs['max_epochs'])
    elif kwargs.get('opt_skip_period', 1) > 1:
        skip_optimization = SkipPeriodicallyPredicate(
            init_length=kwargs['opt_skip_init_length'],
            period=kwargs['opt_skip_period'])
    else:
        skip_optimization = None
    # Profiler
    if kwargs.get('profiler', False):
        profiler = SimpleProfiler()
    else:
        profiler = None
    # Conversion from reward to metric (strictly decreasing) and back
    _map_reward = kwargs.get('map_reward', '1_minus_x')
    if isinstance(_map_reward, str):
        _map_reward_name = _map_reward
        supp_map_reward = {'1_minus_x', 'minus_x'}
        assert _map_reward_name in supp_map_reward, \
            "This factory needs map_reward in {}".format(supp_map_reward)
        _map_reward: MapReward = map_reward(
            const=1.0 if _map_reward_name == '1_minus_x' else 0.0)
    else:
        assert isinstance(_map_reward, MapReward), \
            "map_reward must either be string or of MapReward type"
    if is_hyperband:
        # Note: 'min_reward' is needed only to support the exp-decay
        # surrogate model. If not given, it is assumed to be 0.
        min_reward = kwargs.get('min_reward', 0)
        max_metric_value = _map_reward(min_reward)
    else:
        max_metric_value = None
    # Underlying GP regression model
    gpmodel, model_args = _create_gp_and_model_args(
        hp_ranges_cs=hp_ranges_cs,
        active_metric=DEFAULT_METRIC,
        max_metric_value=max_metric_value,
        is_hyperband=is_hyperband,
        **kwargs)
    debug_log = DebugLogPrinter() if kwargs.get('debug_log', False) else None

    return hp_ranges_cs, random_seed, gpmodel, model_args, profiler, \
           _map_reward, skip_optimization, debug_log


def gp_fifo_searcher_factory(**kwargs) -> GPFIFOSearcher:
    """
    Creates GPFIFOSearcher object, based on kwargs equal to search_options
    passed to and extended by scheduler (see FIFOScheduler).

    Extensions of kwargs by the scheduler:
    - scheduler: Name of scheduler ('fifo', 'hyperband_*')
    - configspace: CS.ConfigurationSpace (or HyperparameterRanges if
      debug_use_hyperparameter_ranges is true)
    Only Hyperband schedulers:
    - resource_attribute: Name of resource (or time) attribute
    - min_epochs: Smallest resource value being rung level
    - max_epochs: Maximum resource value

    :param kwargs: search_options coming from scheduler
    :return: GPFIFOSearcher object

    """
    assert kwargs['scheduler'] in ['fifo', 'local'], \
        "This factory needs scheduler = 'fifo' (instead of '{}')".format(
            kwargs['scheduler'])
    # Common objects
    hp_ranges_cs, random_seed, gpmodel, model_args, profiler, _map_reward, \
    skip_optimization, debug_log = \
        _create_common_objects(**kwargs)

    gp_searcher = GPFIFOSearcher(
        hp_ranges=hp_ranges_cs,
        random_seed=random_seed,
        gpmodel=gpmodel,
        model_args=model_args,
        map_reward=_map_reward,
        acquisition_class=EIAcquisitionFunction,
        skip_optimization=skip_optimization,
        num_initial_candidates=kwargs['num_init_candidates'],
        num_initial_random_choices=kwargs['num_init_random'],
        initial_scoring=kwargs['initial_scoring'],
        profiler=profiler,
        first_is_default=kwargs['first_is_default'],
        debug_log=debug_log,
        cost_metric_name=kwargs.get('cost_metric_name'))
    return gp_searcher


def _resource_for_acquisition(kwargs, hp_ranges_cs):
    _resource_acq = kwargs.get('resource_acq', 'bohb')
    if _resource_acq == 'bohb':
        threshold = kwargs.get(
            'resource_acq_bohb_threshold', hp_ranges_cs.ndarray_size())
        resource_for_acquisition = resource_for_acquisition_bohb(
            threshold=threshold)
    else:
        assert _resource_acq == 'first', \
            "resource_acq must be 'bohb' or 'first'"
        resource_for_acquisition = resource_for_acquisition_first_milestone
    return resource_for_acquisition


def constrained_gp_fifo_searcher_factory(**kwargs) -> ConstrainedGPFIFOSearcher:
    """
    Creates ConstrainedGPFIFOSearcher object, based on kwargs equal to search_options
    passed to and extended by scheduler (see FIFOScheduler).

    :param kwargs: search_options coming from scheduler
    :return: ConstrainedGPFIFOSearcher object

    """
    assert kwargs['scheduler'] == 'fifo', \
        "This factory needs scheduler = 'fifo' (instead of '{}')".format(
            kwargs['scheduler'])
    # Common objects
    hp_ranges_cs, random_seed, gpmodel, model_args, profiler, _map_reward, \
    skip_optimization, debug_log = _create_common_objects(**kwargs)
    initial_scoring = kwargs.get('initial_scoring', 'acq_func')
    # We need two GP models: one for active metric (gpmodel), the other for constraint metric (gpmodel_constraint)
    gpmodel_constraint, model_args_constraint = _create_gp_and_model_args(
        hp_ranges_cs=hp_ranges_cs,
        active_metric=DEFAULT_CONSTRAINT_METRIC,
        **kwargs)
    output_gpmodels = {DEFAULT_METRIC: gpmodel,
                       DEFAULT_CONSTRAINT_METRIC: gpmodel_constraint}
    output_model_args = {DEFAULT_METRIC: model_args,
                         DEFAULT_CONSTRAINT_METRIC: model_args_constraint}
    constrained_gp_searcher = ConstrainedGPFIFOSearcher(
        hp_ranges=hp_ranges_cs,
        random_seed=random_seed,
        output_gpmodels=output_gpmodels,
        output_models_args=output_model_args,
        map_reward=_map_reward,
        acquisition_class=CEIAcquisitionFunction,
        skip_optimization=skip_optimization,
        num_initial_candidates=kwargs['num_init_candidates'],
        num_initial_random_choices=kwargs['num_init_random'],
        initial_scoring=initial_scoring,
        profiler=profiler,
        first_is_default=kwargs['first_is_default'],
        debug_log=debug_log)
    return constrained_gp_searcher


def gp_multifidelity_searcher_factory(**kwargs) -> GPMultiFidelitySearcher:
    """
    Creates GPMultiFidelitySearcher object, based on kwargs equal to search_options
    passed to and extended by scheduler (see HyperbandScheduler).

    :param kwargs: search_options coming from scheduler
    :return: GPMultiFidelitySearcher object

    """
    supp_schedulers = {'hyperband_stopping', 'hyperband_promotion'}
    assert kwargs['scheduler'] in supp_schedulers, \
        "This factory needs scheduler in {} (instead of '{}')".format(
            supp_schedulers, kwargs['scheduler'])
    # Common objects
    hp_ranges_cs, random_seed, gpmodel, model_args, profiler, _map_reward,\
    skip_optimization, debug_log = \
        _create_common_objects(**kwargs)

    resource_for_acquisition = _resource_for_acquisition(kwargs, hp_ranges_cs)
    epoch_range = (kwargs['min_epochs'], kwargs['max_epochs'])
    gp_searcher = GPMultiFidelitySearcher(
        hp_ranges=hp_ranges_cs,
        resource_attr_key=kwargs['resource_attribute'],
        resource_attr_range=epoch_range,
        random_seed=random_seed,
        gpmodel=gpmodel,
        model_args=model_args,
        map_reward=_map_reward,
        acquisition_class=EIAcquisitionFunction,
        resource_for_acquisition=resource_for_acquisition,
        skip_optimization=skip_optimization,
        num_initial_candidates=kwargs['num_init_candidates'],
        num_initial_random_choices=kwargs['num_init_random'],
        initial_scoring=kwargs['initial_scoring'],
        profiler=profiler,
        first_is_default=kwargs['first_is_default'],
        debug_log=debug_log,
        cost_metric_name=kwargs.get('cost_metric_name'))
    return gp_searcher


def _common_defaults(is_hyperband: bool, is_constrained: bool) -> (Set[str], dict, dict):
    mandatory = set()

    default_options = {
        'random_seed': np.random.randint(10000),
        'opt_skip_init_length': 150,
        'opt_skip_period': 1,
        'profiler': False,
        'opt_maxiter': 50,
        'opt_nstarts': 2,
        'opt_warmstart': False,
        'opt_verbose': False,
        'opt_debug_writer': False,
        'num_fantasy_samples': 20,
        'scheduler': 'fifo',
        'num_init_random': DEFAULT_NUM_INITIAL_RANDOM_EVALUATIONS,
        'num_init_candidates': DEFAULT_NUM_INITIAL_CANDIDATES,
        'initial_scoring': DEFAULT_INITIAL_SCORING,
        'first_is_default': True,
        'debug_log': False,
        'cost_metric_name': 'elapsed_time'}
    if is_hyperband:
        default_options['opt_skip_num_max_resource'] = False
        default_options['gp_resource_kernel'] = 'matern52'
        default_options['resource_acq'] = 'bohb'
        default_options['num_init_random'] = 10
    if is_constrained:
        default_options['initial_scoring'] = 'acq_func'

    constraints = {
        'random_seed': Integer(),
        'opt_skip_init_length': Integer(0, None),
        'opt_skip_period': Integer(1, None),
        'profiler': Boolean(),
        'opt_maxiter': Integer(1, None),
        'opt_nstarts': Integer(1, None),
        'opt_warmstart': Boolean(),
        'opt_verbose': Boolean(),
        'opt_debug_writer': Boolean(),
        'num_fantasy_samples': Integer(1, None),
        'num_init_random': Integer(0, None),
        'num_init_candidates': Integer(5, None),
        'initial_scoring': Categorical(
            choices=tuple(SUPPORTED_INITIAL_SCORING)),
        'first_is_default': Boolean(),
        'debug_log': Boolean()}
    if is_hyperband:
        constraints['opt_skip_num_max_resource'] = Boolean()
        constraints['gp_resource_kernel'] = Categorical(choices=(
            'exp-decay-sum', 'exp-decay-combined', 'exp-decay-delta1',
            'matern52', 'matern52-res-warp'))
        constraints['resource_acq'] = Categorical(
            choices=('bohb', 'first'))

    return mandatory, default_options, constraints


def gp_fifo_searcher_defaults() -> (Set[str], dict, dict):
    """
    Returns mandatory, default_options, config_space for
    check_and_merge_defaults to be applied to search_options for
    GPFIFOSearcher.

    :return: (mandatory, default_options, config_space)

    """
    return _common_defaults(is_hyperband=False, is_constrained=False)


def constrained_gp_fifo_searcher_defaults() -> (Set[str], dict, dict):
    """
    Returns mandatory, default_options, config_space for
    check_and_merge_defaults to be applied to search_options for
    ConstrainedGPFIFOSearcher.

    :return: (mandatory, default_options, config_space)

    """
    return _common_defaults(is_hyperband=False, is_constrained=True)


def gp_multifidelity_searcher_defaults() -> (Set[str], dict, dict):
    """
    Returns mandatory, default_options, config_space for
    check_and_merge_defaults to be applied to search_options for
    GPMultiFidelitySearcher.

    :return: (mandatory, default_options, config_space)

    """
    return _common_defaults(is_hyperband=True, is_constrained=False)



import autograd.numpy as anp
from autograd.tracer import getval
from abc import abstractmethod

from ..constants import INITIAL_COVARIANCE_SCALE, INITIAL_INVERSE_BANDWIDTHS, \
    DEFAULT_ENCODING, INVERSE_BANDWIDTHS_LOWER_BOUND, \
    INVERSE_BANDWIDTHS_UPPER_BOUND, COVARIANCE_SCALE_LOWER_BOUND, \
    COVARIANCE_SCALE_UPPER_BOUND, NUMERICAL_JITTER
from ..distribution import Uniform, LogNormal
from ..gluon import Block
from ..gluon_blocks_helpers import encode_unwrap_parameter
from ..mean import MeanFunction
from ..utils import create_encoding, register_parameter

__all__ = ['KernelFunction', 'Matern52']


class KernelFunction(MeanFunction):
    """
    Base class of kernel (or covariance) functions
    """
    def __init__(self, dimension: int, **kwargs):
        """
        :param dimension: Dimensionality of input points after encoding into
            ndarray
        """
        super().__init__(**kwargs)
        self._dimension = dimension

    @property
    def dimension(self):
        """
        :return: Dimension d of input points
        """
        return self._dimension

    @abstractmethod
    def diagonal(self, X):
        """
        :param X: Input data, shape (n, d)
        :return: Diagonal of K(X, X), shape (n,)
        """
        pass

    @abstractmethod
    def diagonal_depends_on_X(self):
        """
        For stationary kernels, diagonal does not depend on X

        :return: Does diagonal(X) depend on X?
        """
        pass

    def _check_input_shape(self, X):
        return anp.reshape(X, (getval(X.shape[0]), self._dimension))


class SquaredDistance(Block):
    """
    Block that is responsible for the computation of matrices of squared
    distances. The distances can possibly be weighted (e.g., ARD
    parametrization). For instance:
        X1 with size (n1,d)
        X2 with size (n2,d)
        inverse_bandwidths with size (1,d)
        results in a matrix of size (n1,n2) with i,j entry equal to
            sum_{k=1}^d (X1[i,k] - X2[j,k])^2 * inverse_bandwidths[k]^2

    if ARD == False, inverse_bandwidths is equal to a scalar broadcast to the
    d components (with d=dimension, i.e., the number of features in X)
    otherwise, inverse_bandwidths is (1,d)
    """
    def __init__(self, dimension, ARD=False, encoding_type=DEFAULT_ENCODING, **kwargs):
        super().__init__(**kwargs)
        self.ARD = ARD
        inverse_bandwidths_dimension = 1 if not ARD else dimension
        self.encoding = create_encoding(
            encoding_type, INITIAL_INVERSE_BANDWIDTHS,
            INVERSE_BANDWIDTHS_LOWER_BOUND, INVERSE_BANDWIDTHS_UPPER_BOUND,
            inverse_bandwidths_dimension,
            Uniform(INVERSE_BANDWIDTHS_LOWER_BOUND,
                    INVERSE_BANDWIDTHS_UPPER_BOUND))
        
        with self.name_scope():
            self.inverse_bandwidths_internal = register_parameter(
                self.params, 'inverse_bandwidths', self.encoding,
                shape=(inverse_bandwidths_dimension,))

    def _inverse_bandwidths(self):
        return encode_unwrap_parameter(
            self.inverse_bandwidths_internal, self.encoding)

    def forward(self, X1, X2):
        """
        Actual computation of the matrix of squared distances (see details above)

        :param X1: input data of size (n1,d)
        :param X2: input data of size (n2,d)
        :param inverse_bandwidths_internal: self.inverse_bandwidths_internal
        """
        # In case inverse_bandwidths if of size (1, dimension), dimension>1,
        # ARD is handled by broadcasting
        inverse_bandwidths = anp.reshape(self._inverse_bandwidths(), (1, -1))

        if X2 is X1:
            X1_scaled = anp.multiply(X1, inverse_bandwidths)
            D = -2.0 * anp.dot(X1_scaled, anp.transpose(X1_scaled))
            X1_squared_norm = anp.sum(anp.square(X1_scaled), axis=1)
            D = D + anp.reshape(X1_squared_norm, (1, -1))
            D = D + anp.reshape(X1_squared_norm, (-1, 1))
        else:
            X1_scaled = anp.multiply(X1, inverse_bandwidths)
            X2_scaled = anp.multiply(X2, inverse_bandwidths)
            X1_squared_norm = anp.sum(anp.square(X1_scaled), axis=1)
            X2_squared_norm = anp.sum(anp.square(X2_scaled), axis=1)
            D = -2.0 * anp.matmul(
                X1_scaled, anp.transpose(X2_scaled))
            D = D + anp.reshape(X1_squared_norm, (-1, 1))
            D = D + anp.reshape(X2_squared_norm, (1, -1))

        return anp.abs(D)

    def get_params(self):
        """
        Parameter keys are inv_bw<k> if dimension > 1, and inv_bw if
        dimension == 1.
        """
        inverse_bandwidths = anp.reshape(self._inverse_bandwidths(), (-1,))
        if inverse_bandwidths.size == 1:
            return {'inv_bw': inverse_bandwidths[0]}
        else:
            return {
                'inv_bw{}'.format(k): inverse_bandwidths[k]
                for k in range(inverse_bandwidths.size)}

    def set_params(self, param_dict):
        dimension = self.encoding.dimension
        if dimension == 1:
            inverse_bandwidths = [param_dict['inv_bw']]
        else:
            keys = ['inv_bw{}'.format(k) for k in range(dimension)]
            for k in keys:
                assert k in param_dict, "'{}' not in param_dict = {}".format(k, param_dict)
            inverse_bandwidths = [param_dict[k] for k in keys]
        self.encoding.set(self.inverse_bandwidths_internal, inverse_bandwidths)
        

class Matern52(KernelFunction):
    """
    Block that is responsible for the computation of Matern52 kernel
    matrices. For instance:
        X1 with size (n1,d)
        X2 with size (n2,d)
    results in a matrix of size (n1,n2) with i,j entry equal to the
    Matern52 kernel at (X1[i,:], X2[j,:]).

    If ARD == False, inverse_bandwidths is equal to a scalar broadcast to the
    d components (with d=dimension, i.e., the number of features in X)
    otherwise (ARD == True), inverse_bandwidths is (1,d)
    """
    def __init__(self, dimension, ARD=False, encoding_type=DEFAULT_ENCODING,
                 **kwargs):
        super(Matern52, self).__init__(dimension, **kwargs)
        self.encoding = create_encoding(
            encoding_type, INITIAL_COVARIANCE_SCALE,
            COVARIANCE_SCALE_LOWER_BOUND, COVARIANCE_SCALE_UPPER_BOUND, 1,
            LogNormal(0.0, 1.0))
        self.ARD = ARD
        self.squared_distance = SquaredDistance(
            dimension=dimension, ARD=ARD, encoding_type=encoding_type)
            
        with self.name_scope():
            self.covariance_scale_internal = register_parameter(
                self.params, 'covariance_scale', self.encoding)

    def _covariance_scale(self):
        return encode_unwrap_parameter(
            self.covariance_scale_internal, self.encoding)

    def forward(self, X1, X2):
        """
        Actual computation of the Matern52 kernel matrix (see details above)
        See http://www.gaussianprocess.org/gpml/chapters/RW.pdf,
        equation (4.17)

        :param X1: input data of size (n1, d)
        :param X2: input data of size (n2, d)
        """
        covariance_scale = self._covariance_scale()
        X1 = self._check_input_shape(X1)
        if X2 is not X1:
            X2 = self._check_input_shape(X2)
        D = self.squared_distance(X1, X2)
        # Using the plain np.sqrt is numerically unstable for D ~ 0
        # (non-differentiability)
        # that's why we add NUMERICAL_JITTER
        B = anp.sqrt(5.0 * D + NUMERICAL_JITTER)
        K = anp.multiply(
            (1.0 + B + 5.0 / 3.0 * D) * anp.exp(-B), covariance_scale)

        return K

    def diagonal(self, X):
        X = self._check_input_shape(X)
        covariance_scale = self._covariance_scale()
        covariance_scale_times_ones = anp.multiply(
            anp.ones((getval(X.shape[0]), 1)), covariance_scale)
        
        return anp.reshape(covariance_scale_times_ones, (-1,))

    def diagonal_depends_on_X(self):
        return False

    def param_encoding_pairs(self):
        return [
            (self.covariance_scale_internal, self.encoding),
            (self.squared_distance.inverse_bandwidths_internal,
             self.squared_distance.encoding)
        ]
        
    def get_covariance_scale(self):
        return self._covariance_scale()[0]

    def set_covariance_scale(self, covariance_scale):
        self.encoding.set(self.covariance_scale_internal, covariance_scale)

    def get_params(self):
        result = self.squared_distance.get_params()
        result['covariance_scale'] = self.get_covariance_scale()
        return result

    def set_params(self, param_dict):
        self.squared_distance.set_params(param_dict)
        self.set_covariance_scale(param_dict['covariance_scale'])



from typing import List, Optional
import numpy as np
import logging

from ..datatypes.tuning_job_state import TuningJobState
from ..tuning_algorithms.base_classes import SurrogateModel
from ..utils.debug_log import DebugLogPrinter

logger = logging.getLogger(__name__)


class BaseSurrogateModel(SurrogateModel):
    """
    Base class for (most) SurrogateModel implementations, provides common code

    """
    def __init__(
            self, state: TuningJobState, active_metric: str, random_seed: int,
            debug_log: Optional[DebugLogPrinter] = None):
        super().__init__(state, active_metric, random_seed)
        self._current_best = None
        self._debug_log = debug_log

    def predict_mean_current_candidates(self) -> List[np.ndarray]:
        """
        Returns the predictive mean (signal with key 'mean') at all current candidate
        locations (both state.candidate_evaluations and state.pending_evaluations).

        If the hyperparameters of the surrogate model are being optimized (e.g.,
        by empirical Bayes), the returned list has length 1. If its
        hyperparameters are averaged over by MCMC, the returned list has one
        entry per MCMC sample.

        :return: List of predictive means
        """
        candidates = [
            x.candidate for x in self.state.candidate_evaluations] + \
                     self.state.pending_candidates
        candidates = self.current_best_filter_candidates(candidates)
        assert len(candidates) > 0, \
            "Cannot predict means at current candidates with no candidates at all"

        inputs = self.state.hp_ranges.to_ndarray_matrix(candidates)

        all_means = []
        # Loop over MCMC samples (if any)
        for prediction in self.predict(inputs):
            means = prediction['mean']
            if means.ndim == 1:  # In case of no fantasizing
                means = means.reshape((-1, 1))
            all_means.append(means)
        return all_means

    def current_best(self) -> List[np.ndarray]:
        if self._current_best is None:
            all_means = self.predict_mean_current_candidates()
            result = [np.min(means, axis=0)
                      for means in all_means]
            self._current_best = result
        return self._current_best

    def current_best_filter_candidates(self, candidates):
        """
        In some subclasses, 'current_best' is not computed over all (observed
        and pending) candidates: they need to implement this filter.

        """
        return candidates  # Default: No filtering
